/*
** k_linux.c - Linux 0.99.13q or later kernel access functions
**
** Copyright (c) 1997-1998 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#include <string.h>
#include <stdio.h>
#include <syslog.h>
#include <errno.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <linux/types.h>
#include <linux/netlink.h>
#include <linux/inet_diag.h>

#include "pidentd.h"

struct kainfo
{
    int nlfd;
    __u32 seq;
    FILE *proc_net_tcp;
};

/*
** Make sure we are running on a supported OS version
*/
int
ka_init(void)
{
    return 0; /* We always succeed */
}


int
ka_open(void **misc)
{
    struct kainfo *kp;

    kp = s_malloc(sizeof(*kp));

    kp->seq = 0;
    kp->nlfd = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_INET_DIAG);
    if (kp->nlfd >= 0)
	goto out;
    
    syslog(LOG_INFO, "netlink failed, fallback to /proc/net/tcp: %m");
    kp->proc_net_tcp = fopen("/proc/net/tcp", "r");
    
    if (kp->proc_net_tcp == NULL)
    {
	syslog(LOG_ERR, "fopen(\"/proc/net/tcp\", \"r\"): %m");
	return -1;
    }

out:
    *misc = (void *) kp;
    return 0;
}


static int
netlink_lookup(struct kainfo *kip, struct kernel *kp)
{
    int status;
    int i;
    struct {
	struct nlmsghdr nlh;
	union {
		struct inet_diag_req req;
		struct inet_diag_msg rsp;
	} u;
    } buf;
    struct sockaddr_nl addr;

    memset(&buf, 0, sizeof(buf));
    buf.nlh.nlmsg_len = NLMSG_ALIGN(NLMSG_LENGTH(sizeof(buf.u.req)));
    buf.nlh.nlmsg_type = TCPDIAG_GETSOCK;
    buf.nlh.nlmsg_flags = NLM_F_REQUEST;
    buf.nlh.nlmsg_seq = ++kip->seq;
    buf.u.req.idiag_family = kp->remote.sg_addr._sg_sa.sa_family;

    switch (buf.u.req.idiag_family) {

    case AF_INET:
	buf.u.req.id.idiag_dport = kp->remote.sg_addr._sg_sin.sin_port;
	buf.u.req.id.idiag_sport = kp->local.sg_addr._sg_sin.sin_port;
	buf.u.req.id.idiag_dst[0] = kp->remote.sg_addr._sg_sin.sin_addr.s_addr;
	buf.u.req.id.idiag_src[0] = kp->local.sg_addr._sg_sin.sin_addr.s_addr;
	break;

    case AF_INET6:
	buf.u.req.id.idiag_dport = kp->remote.sg_addr._sg_sin6.sin6_port;
	buf.u.req.id.idiag_sport = kp->local.sg_addr._sg_sin6.sin6_port;

	for (i = 0; i < 4; i++) {
	    buf.u.req.id.idiag_dst[i] =
		kp->remote.sg_addr._sg_sin6.sin6_addr.__in6_u.__u6_addr32[i];
	    buf.u.req.id.idiag_src[i] =
		kp->local.sg_addr._sg_sin6.sin6_addr.__in6_u.__u6_addr32[i];
	}
	break;

    default:
	syslog(LOG_ERR, "netlink_lookup: unknown address family: %d",
	       buf.u.req.idiag_family);
	return 3;
    }
    buf.u.req.id.idiag_cookie[0] = INET_DIAG_NOCOOKIE;
    buf.u.req.id.idiag_cookie[1] = INET_DIAG_NOCOOKIE;

    status = write(kip->nlfd, &buf, buf.nlh.nlmsg_len);
    if (status < 0) {
	syslog(LOG_ERR, "netlink_lookup: write failed: %m");
	return 3;
    }

    do {
	socklen_t alen = sizeof(addr);
	status = recvfrom(kip->nlfd, &buf, sizeof(buf), 0,
	 			(void *)&addr, &alen);
	if (status < 0) {
	    if (errno == ENOBUFS)
		return -1;
	    syslog(LOG_ERR, "netlink_lookup: recvfrom failed: %m");
	    return 3;
	}
    } while (addr.nl_pid || buf.nlh.nlmsg_seq != kip->seq);

    if (buf.nlh.nlmsg_type != TCPDIAG_GETSOCK)
	return 0;
    if (buf.u.rsp.idiag_state != TCP_ESTABLISHED)
	return 0;

    kp->euid = buf.u.rsp.idiag_uid;
    return 1;
}


int 
ka_lookup(void *vp, struct kernel *kp)
{
    FILE *fp;
    long dummy;
    char buf[512];
    long r_laddr, r_raddr, myladdr, myraddr;
    int r_lport, r_rport, mylport, myrport;
    int euid;
    int nra;
    int status;
    unsigned long ino;
    struct kainfo *kip;
    
/*
 * PSz 11 Dec 02
 * 
 * We have observed Debian identd (on a fairly busy dual-CPU machine)
 * sometimes reporting 'root' for connections belonging to "real" users.
 * File /proc/net/tcp is written mostly in routine get__sock in
 * /usr/src/kernel-source-2.2.19/net/ipv4/proc.c :
 * 
 * 	sprintf(tmpbuf, "%4d: %08lX:%04X %08lX:%04X"
 * 		" %02X %08X:%08X %02X:%08lX %08X %5d %8d %lu",
 * 		i, src, srcp, dest, destp, sp->state, 
 * 		(tw_bucket ?
 * 		 0 :
 * 		 (format == 0) ?
 * 		 tp->write_seq-tp->snd_una : atomic_read(&sp->wmem_alloc)),
 * 		(tw_bucket ?
 * 		 0 :
 * 		 (format == 0) ?
 * 		 tp->rcv_nxt-tp->copied_seq: atomic_read(&sp->rmem_alloc)),
 * 		timer_active, timer_expires-jiffies,
 * 		(tw_bucket ? 0 : tp->retransmits),
 * 		(!tw_bucket && sp->socket) ? sp->socket->inode->i_uid : 0,
 * 		(!tw_bucket && timer_active) ? sp->timeout : 0,
 * 		(!tw_bucket && sp->socket) ? sp->socket->inode->i_ino : 0);
 * 
 * Should we skip lines with just ino, or both uid and ino, zero?
 */


    kip = (struct kainfo *)vp;
    kp->ruid = NO_UID;

    if (kip->nlfd >= 0)
	return netlink_lookup(kip, kp);

    r_rport = ntohs(kp->remote.sg_addr._sg_sin.sin_port);
    r_lport = ntohs(kp->local.sg_addr._sg_sin.sin_port);
    r_raddr = kp->remote.sg_addr._sg_sin.sin_addr.s_addr;
    r_laddr = kp->local.sg_addr._sg_sin.sin_addr.s_addr;

    fp = kip->proc_net_tcp;

    rewind(fp);

    /* eat header */
    if (fgets(buf, sizeof(buf)-1,fp) == NULL)
	return -1;

    status = 0;
    while (fgets(buf, sizeof(buf)-1, fp) != NULL)
    {
	nra = sscanf(buf, "%d: %lX:%x %lX:%x %x %lX:%lX %x:%lX %lx %d %ld %lu",
		     &dummy, &myladdr, &mylport, &myraddr, &myrport,
		     &dummy, &dummy, &dummy, &dummy, &dummy, &dummy,
		     &euid, &dummy, &ino);
	if (nra >= 12)
	{
	    if (myladdr == r_laddr && mylport == r_lport)
	    {
		if (myraddr != r_raddr || myrport != r_rport || !ino)
		    continue;
		if (nra >= 14 && euid == 0 && ino == 0) {
		  /*
		   * Both uid and ino are zero: not even a socket?
		   * Skip (continue), probably fail; or fail (break)
		   * straight away? Hopefully we retry later.
		   */
		  continue;
		}
		kp->euid = euid;
		status = 1;
	    }
	}
    }

    return status;
}

